mod configuration;

use std::{
    ffi::{c_int, c_void, CString},
    sync::atomic::{AtomicBool, AtomicU16, Ordering},
};

use chuniio_tasoller_common::{create_input_shared_memory, log::init_logger};
use lazy_static::lazy_static;
use winapi::{
    shared::{
        minwindef::{BOOL, DWORD, HINSTANCE, LPVOID, TRUE},
        winerror::{HRESULT, S_OK},
    },
    um::{winbase::GetPrivateProfileIntA, winnt::DLL_PROCESS_ATTACH, winuser::GetAsyncKeyState},
};

use crate::configuration::Configuration;

static COIN_COUNT: AtomicU16 = AtomicU16::new(0);
static COIN_PRESSED: AtomicBool = AtomicBool::new(false);

lazy_static! {
    static ref CONFIGURATION: Configuration = {
        let io3 = CString::new("io3").unwrap();
        let test = CString::new("test").unwrap();
        let service = CString::new("service").unwrap();
        let coin = CString::new("coin").unwrap();
        let cfg_file = CString::new(".\\segatools.ini").unwrap();

        unsafe {
            Configuration {
                test_key: GetPrivateProfileIntA(
                    io3.as_ptr(),
                    test.as_ptr(),
                    0x31,
                    cfg_file.as_ptr(),
                ),
                service_key: GetPrivateProfileIntA(
                    io3.as_ptr(),
                    service.as_ptr(),
                    0x32,
                    cfg_file.as_ptr(),
                ),
                coin_key: GetPrivateProfileIntA(
                    io3.as_ptr(),
                    coin.as_ptr(),
                    0x33,
                    cfg_file.as_ptr(),
                ),
            }
        }
    };
}

#[no_mangle]
extern "system" fn DllMain(_dll_module: HINSTANCE, call_reason: DWORD, _reserved: LPVOID) -> BOOL {
    match call_reason {
        DLL_PROCESS_ATTACH => {
            init_logger();
        }
        _ => {}
    }

    TRUE
}

#[no_mangle]
pub extern "C" fn chuni_io_get_api_version() -> u16 {
    0x0102
}

#[no_mangle]
pub extern "C" fn chuni_io_jvs_init() -> HRESULT {
    S_OK
}

#[no_mangle]
pub unsafe extern "C" fn chuni_io_jvs_poll(opbtn: *mut u8, beams: *mut u8) {
    let Ok(input_shmem) = create_input_shared_memory(false) else {
        return;
    };
    let input = input_shmem.as_slice();
    let bit = input[3];

    if bit & (1 << 6) != 0 {
        *opbtn |= 0x1;
    }

    // fn2
    if bit & (1 << 7) != 0 {
        *opbtn |= 0x2;
    }

    for i in 0..6 {
        if bit & (1 << i) != 0 {
            *beams |= 1 << i;
        }
    }
}

#[no_mangle]
pub unsafe extern "C" fn chuni_io_jvs_read_coin_counter(out: *mut u16) {
    if out.is_null() {
        return;
    }

    if GetAsyncKeyState(CONFIGURATION.coin_key as c_int) != 0 {
        let coin_previously_pressed = COIN_PRESSED
            .fetch_update(Ordering::Relaxed, Ordering::Relaxed, |_| Some(true))
            .unwrap();
        if !coin_previously_pressed {
            COIN_COUNT.fetch_add(1, Ordering::Relaxed);
        }
    } else {
        COIN_PRESSED.store(false, Ordering::Relaxed);
    }

    *out = COIN_COUNT.load(Ordering::Relaxed);
}

// Stubs. These are handled by chusanApp
#[no_mangle]
pub extern "C" fn chuni_io_slider_init() -> HRESULT {
    S_OK
}

#[no_mangle]
pub extern "C" fn chuni_io_slider_start(_callback: *const c_void) {}

#[no_mangle]
pub extern "C" fn chuni_io_slider_set_leds(_rgb: *mut u8) {}

#[no_mangle]
pub extern "C" fn chuni_io_slider_stop() {}

#[no_mangle]
pub extern "C" fn chuni_io_led_init() -> HRESULT {
    S_OK
}

#[no_mangle]
pub extern "C" fn chuni_io_led_set_colors(_board: u8, _data: *const u8) {}
